import logging

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from .package_data import get_path_of_data_file


class LogFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno != self.__level


_console_formatter = logging.Formatter(
    ' %(message)s',
    datefmt="%H:%M:%S",
)


mytheme = Theme().read(get_path_of_data_file("log_theme.ini"))
console = Console(theme=mytheme)

ronswanson_console_log_handler = RichHandler(
    level="INFO", rich_tracebacks=True, markup=True, console=console
)
ronswanson_console_log_handler.setFormatter(_console_formatter)

warning_filter = LogFilter(logging.WARNING)


def silence_warnings():
    """
    supress warning messages in console and file usr logs
    """

    ronswanson_console_log_handler.addFilter(warning_filter)


def activate_warnings():
    """
    supress warning messages in console and file usr logs
    """

    ronswanson_console_log_handler.removeFilter(warning_filter)


def update_logging_level(level):

    ronswanson_console_log_handler.setLevel(level)


def setup_logger(name):

    # A logger with name name will be created
    # and then add it to the print stream
    log = logging.getLogger(name)

    # this must be set to allow debug messages through
    log.setLevel(logging.DEBUG)

    # add the handlers
    log.addHandler(ronswanson_console_log_handler)

    # we do not want to duplicate teh messages in the parents
    log.propagate = False

    return log



from typing import Dict

import numpy as np
from astromodels import Band

from .simulation import Simulation


class BandSimulation(Simulation):
    def __init__(
        self,
        simulation_id: int,
        parameter_set: Dict[str, float],
        energy_grid: np.ndarray,
        out_file: str,
    ) -> None:
        super().__init__(simulation_id, parameter_set, energy_grid, out_file)

    def _run_call(self) -> np.ndarray:

        b = Band(
            K=1,
            alpha=self._parameter_set["alpha"],
            beta=self._parameter_set["beta"],
            xp=self._parameter_set["epeak"],
        )

        return b(self._energy_grid)



import time
from abc import ABCMeta, abstractmethod
from typing import Dict

import h5py
import numpy as np

from .utils.logging import setup_logger

log = setup_logger(__name__)


def file_is_open(file_name: str):

    try:
        tmp = h5py.File(file_name, "a")

        return False, tmp

    except OSError:

        return True, None


class Simulation(metaclass=ABCMeta):
    def __init__(
        self,
        simulation_id: int,
        parameter_set: Dict[str, float],
        energy_grid: np.ndarray,
        out_file: str,
    ) -> None:

        """
        Generic simulation class

        :param simulation_id:
        :type simulation_id: int
        :param parameter_set:
        :type parameter_set: Dict[str, float]
        :param energy_grid:
        :type energy_grid: np.ndarray
        :param out_file:
        :type out_file: str
        :returns:

        """
        self._out_file: str = out_file
        self._parameter_set: Dict[str, float] = parameter_set
        self._simulation_id: int = simulation_id
        self._energy_grid: np.ndarray = energy_grid

    def run(self) -> None:

        """
        run this simulation
        :returns:

        """

        # check if this already exists

        params = np.array(list(self._parameter_set.values()))

        run_flag: bool = True

        while True:

            test, f = file_is_open(self._out_file)

            if test:

                # the file is already open so wait

                log.debug(
                    f"simulation {self._simulation_id} is waiting on file to be closed"
                )

                time.sleep(np.random.randint(3, 5))

            else:

                if "parameters" not in f.keys():

                    # ok, this is a brand new file
                    log.debug("New database file")
                    break

                else:

                    for k, v in f["parameters"].items():

                        if np.alltrue(v[()] == params):

                            # this parameter set exists

                            run_flag = False

                            log.debug(f"parameters {v[()]} already exist!")

                            break

                f.close()

                break

        if not run_flag:

            log.debug(f"simulation {self._simulation_id} not running")

            return

        # run the simulation

        log.debug(f"simulation {self._simulation_id} is now running")

        output: np.ndarray = self._run_call()

        while True:

            test, f = file_is_open(self._out_file)

            if test:

                # the file is already open so wait

                log.debug(
                    f"simulation {self._simulation_id} is waiting on file to be closed"
                )

                time.sleep(np.random.randint(3, 5))

            else:

                log.debug(f"simulation {self._simulation_id} is storing")

                # store the parameter names

                if "parameter_names" not in f.keys():

                    p_name_group = f.create_group("parameter_names")

                    for i, name in enumerate(list(self._parameter_set.keys())):

                        p_name_group.attrs[f"par{i}"] = name

                # store the energy grid

                if "energy_grid" not in f.keys():

                    f.create_dataset(
                        "energy_grid",
                        data=self._energy_grid,
                        compression="gzip",
                    )

                if "parameters" not in f.keys():

                    f.create_group("parameters")

                param_group: h5py.Group = f["parameters"]

                number_of_entries: int = len(param_group.keys())

                new_key: int = number_of_entries

                params = np.array(list(self._parameter_set.values()))

                param_group.create_dataset(
                    f"{new_key}", data=params, compression="gzip"
                )

                if "values" not in f.keys():

                    f.create_group("values")

                values_group: h5py.Group = f["values"]

                values_group.create_dataset(
                    f"{new_key}", data=output, compression="gzip"
                )

                f.close()

                break

    @abstractmethod
    def _run_call(self) -> np.ndarray:

        log.error("Attempting to use base class")

        raise RuntimeError()



from collections import OrderedDict
from typing import Dict, List, Optional

import h5py
import numpy as np
from astromodels import TemplateModel, TemplateModelFactory

from .simulation_builder import Parameter, ParameterGrid
from .utils.logging import setup_logger

log = setup_logger(__name__)


class Database:
    def __init__(
        self,
        grid_points: Dict[str, np.ndarray],
        parameter_names: List[str],
        energy_grid: np.ndarray,
        values: Dict[str, np.ndarray],
    ) -> None:
        """
        Databse of parameters and simulated values

        :param grid_points:
        :type grid_points: Dict[str, np.ndarray]
        :param parameter_names:
        :type parameter_names: List[str]
        :param energy_grid:
        :type energy_grid: np.ndarray
        :param values:
        :type values: Dict[str, np.ndarray]
        :returns:

        """

        self._n_entries = len(values)

        self._n_parameters: int = len(parameter_names)

        self._parameter_names: List[str] = parameter_names

        self._energy_grid: np.ndarray = energy_grid

        # self._values: Dict[str, np.ndarray] = values
        # self._grid_points: Dict[str, np.ndarray] = grid_points

        self._values: np.ndarray = np.empty(
            (self._n_entries, len(self._energy_grid))
        )

        self._grid_points: np.ndarray = np.empty(
            (self._n_entries, self._n_parameters)
        )

        self._extract_parameter_values(values, grid_points)

    @property
    def n_entries(self) -> int:
        return self._n_entries

    @property
    def n_parameters(self) -> int:
        return self._n_parameters

    @property
    def parameter_ranges(self) -> Dict[str, np.ndarray]:
        return self._parameter_ranges

    def _extract_parameter_values(self, values, grid_points):

        """
        extract the values and parameter ranges for
        the tables

        :param values:
        :type values:
        :param grid_points:
        :type grid_points:
        :returns:

        """

        # extract the values

        for i in range(self._n_entries):

            self._values[i] = values[f"{i}"]
            self._grid_points[i] = grid_points[f"{i}"]

        self._parameter_ranges = OrderedDict()

        for i, par in enumerate(self._parameter_names):

            self._parameter_ranges[par] = np.sort(
                np.unique(self._grid_points[:, i])
            )

    def _sub_selection(
        self,
        paramater_name: str,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> np.ndarray:

        if paramater_name not in self._parameter_names:

            log.error(f"{paramater_name} is not in {self._parameter_names}")

            raise AssertionError()

        par_idx: int = self._parameter_names.index(paramater_name)

        selection = np.ones(self._n_entries, dtype=bool)

        if vmin is not None:

            selection = (self._grid_points[..., par_idx] >= vmin) & selection

        if vmax is not None:

            selection = (self._grid_points[..., par_idx] <= vmax) & selection

        return selection

    def _parameter_sub_selection(
        self,
        paramater_name: str,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> np.ndarray:

        if paramater_name not in self._parameter_names:

            log.error(f"{paramater_name} is not in {self._parameter_names}")

            raise AssertionError()

        par_range = self._parameter_ranges[paramater_name]

        selection = np.ones(len(par_range), dtype=bool)

        if vmin is not None:

            selection = (par_range >= vmin) & selection

        if vmax is not None:

            selection = (par_range <= vmax) & selection

        return selection

    @classmethod
    def from_file(cls, file_name: str) -> "Database":

        values = {}
        parameters = {}

        with h5py.File(file_name, "r") as f:

            energy_grid = f['energy_grid'][()]

            values_grp = f["values"]

            parameters_grp = f["parameters"]

            par_name_grp = f["parameter_names"]

            parameter_names = [
                par_name_grp.attrs[f"par{i}"]
                for i in range(len(par_name_grp.attrs))
            ]
            for key in values_grp.keys():

                values[key] = values_grp[key][()]
                parameters[key] = parameters_grp[key][()]

        return cls(parameters, parameter_names, energy_grid, values)

    def to_3ml(
        self, name: str, desc: str, overwrite: bool = False, **kwargs
    ) -> TemplateModel:

        """
        construct a table model from the database.
        parameter sub-selections are passed as kwargs of
        dictionaries:

        selections = dict(param1=dict(vmin=1, vmax=2))

        :param name:
        :type name: str
        :param desc:
        :type desc: str
        :param overwrite:
        :type overwrite: bool
        :returns:

        """
        selection = np.ones(self._n_entries, dtype=bool)

        parameter_selection = {}

        for k, v in self._parameter_ranges.items():

            parameter_selection[k] = np.ones(len(v), dtype=bool)

        for k, v in kwargs.items():

            if k in self._parameter_names:

                vmin = None
                vmax = None

                if 'vmin' in v:

                    vmin = v['vmin']

                if 'vmax' in v:

                    vmax = v['vmax']

                selection = selection & self._sub_selection(k, vmin, vmax)

                parameter_selection[k] = parameter_selection[
                    k
                ] & self._parameter_sub_selection(k, vmin, vmax)

        # sub selections if any

        sub_grid = self._grid_points[selection, ...]
        sub_values = self._values[selection, ...]

        sub_parameter_ranges = {}

        for k, v in parameter_selection.items():

            sub_parameter_ranges[k] = self._parameter_ranges[k][
                parameter_selection[k]
            ]

        tmf = TemplateModelFactory(
            name, desc, self._energy_grid, self._parameter_names
        )

        for k, v in sub_parameter_ranges.items():

            tmf.define_parameter_grid(k, sub_parameter_ranges[k])

        for i in range(len(sub_values)):

            ### DO NOT SORT

            tmf.add_interpolation_data(
                sub_values[i],
                **{k: v for k, v in zip(self._parameter_names, sub_grid[i])},
            )

        tmf.save_data(overwrite=True)

        return TemplateModel(name)


__all__ = ["Database"]



# from typing import Any, Dict

# import h5py
# import numpy as np


# def recursively_save_dict_contents_to_group(h5file, path, dic: Dict[Any, Any]):
#     """
#     Save dict to HDF5.

#     :param h5file: HDF5 file
#     :param path: Path inside file
#     :param dic: Dictionary to save
#     :type dic: Dict[Any, Any]
#     """
#     for key, item in dic.items():

#         if isinstance(item, list):

#             is_ok = True

#             for i in item:

#                 if isinstance(i, str):

#                     pass

#                 else:

#                     is_ok = False

#             if is_ok:

#                 # now create a dict

#                 tmp = {}

#                 for i in item:

#                     tmp[i] = {}

#                 item = tmp

#             else:

#                 raise ValueError("Cannot save %s type" % type(item))

#         if isinstance(
#             item, (np.ndarray, np.int64, np.float64, str, bytes, float, int)
#         ):

#             h5file[path + "/" + key] = item

#         elif isinstance(item, dict):

#             if len(item) == 0:

#                 h5file[path + "/" + key] = "FILL_VALUE"

#             recursively_save_dict_contents_to_group(
#                 h5file, path + "/" + key + "/", item
#             )
#         else:
#             raise ValueError("Cannot save %s type" % type(item))


# def recursively_load_dict_contents_from_group(h5file, path):
#     """
#     Load files from hdf5.

#     :param h5file: HDF5 file
#     :param path: Path in file
#     :returns: A dictionary
#     :rtype: Dict[Any, Any]
#     """
#     ans = {}

#     for key, item in h5file[path].items():

#         if isinstance(item, h5py._hl.dataset.Dataset):
#             ans[key] = item[()]

#             try:

#                 ans[key] = ans[key].decode()

#                 if ans[key] == "FILL_VALUE":

#                     ans[key] = {}

#             except:

#                 pass

#         elif isinstance(item, h5py._hl.group.Group):
#             ans[key] = recursively_load_dict_contents_from_group(
#                 h5file, path + "/" + key + "/"
#             )
#     return ans



import itertools
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import h5py
import numpy as np
import yaml

from .script_generator import PythonGenerator, SLURMGenerator
from .simulation import Simulation

# from .utils.hdf5_utils import recursively_save_dict_contents_to_group
from .utils.logging import setup_logger

log = setup_logger(__name__)


@dataclass(frozen=True)
class EnergyGrid:

    vmin: Optional[float] = None
    vmax: Optional[float] = None
    scale: Optional[str] = None
    n_points: Optional[int] = None
    values: Optional[np.ndarray] = None
    custom: bool = False

    def __post_init__(self):

        if not self.custom:

            # we will build a grid

            if (self.vmin is None) or (self.vmax is None):

                log.error("non-custom grids must include vmin and vmax")

                raise AssertionError

            if self.scale is None:

                log.error(
                    "non-custom grids must include scale 'log' or 'linear'"
                )

                raise AssertionError

            else:

                if self.scale not in ['log', 'linear']:

                    log.error(
                        "non-custom grids must include scale 'log' or 'linear'"
                    )

                    raise AssertionError

            if self.n_points is None:

                log.error("non-custom grids must include n_points")

                raise AssertionError

        else:

            if self.values is None:

                log.error("custom grids must include values")

                raise AssertionError

    @property
    def grid(self) -> np.ndarray:

        if self.custom:

            return self.values

        else:

            if self.scale.lower() == 'log':

                return np.geomspace(self.vmin, self.vmax, self.n_points)

            else:

                return np.linspace(self.vmin, self.vmax, self.n_points)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Parameter":

        inputs = {}
        inputs["custom"] = d["custom"]

        if d["custom"]:

            inputs["values"] = np.array(d["values"])

        else:

            inputs["vmin"] = d["vmin"]
            inputs["vmax"] = d["vmax"]
            inputs["scale"] = d["scale"]
            inputs["n_points"] = d["n_points"]

        return cls(**inputs)

    def to_dict(self) -> Dict[str, Any]:

        out = dict(custom=self.custom)

        if self.custom:

            out["values"] = self.values.tolist()

        else:

            out["vmin"] = self.vmin
            out["vmax"] = self.vmax
            out["scale"] = self.scale
            out["n_points"] = self.n_points

        return out


@dataclass(frozen=True)
class Parameter:

    name: str
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    scale: Optional[str] = None
    n_points: Optional[int] = None
    values: Optional[np.ndarray] = None
    custom: bool = False
    #   grid: np.ndarray = field(init=False)

    def __post_init__(self):

        if not self.custom:

            # we will build a grid

            if (self.vmin is None) or (self.vmax is None):

                log.error("non-custom grids must include vmin and vmax")

                raise AssertionError

            if self.scale is None:

                log.error(
                    "non-custom grids must include scale 'log' or 'linear'"
                )

                raise AssertionError

            else:

                if self.scale not in ['log', 'linear']:

                    log.error(
                        "non-custom grids must include scale 'log' or 'linear'"
                    )

                    raise AssertionError

            if self.n_points is None:

                log.error("non-custom grids must include n_points")

                raise AssertionError

        else:

            if self.values is None:

                log.error("custom grids must include values")

                raise AssertionError

    @property
    def grid(self) -> np.ndarray:

        if self.custom:

            return self.values

        else:

            if self.scale.lower() == 'log':

                return np.geomspace(self.vmin, self.vmax, self.n_points)

            else:

                return np.linspace(self.vmin, self.vmax, self.n_points)

    @classmethod
    def from_dict(cls, name: str, d: Dict[str, Any]) -> "Parameter":

        inputs = {}
        inputs["custom"] = d["custom"]

        if d["custom"]:

            inputs["values"] = np.array(d["values"])

        else:

            inputs["vmin"] = d["vmin"]
            inputs["vmax"] = d["vmax"]
            inputs["scale"] = d["scale"]
            inputs["n_points"] = d["n_points"]

        return cls(name, **inputs)

    def to_dict(self) -> Dict[str, Any]:

        out = dict(custom=self.custom)

        if self.custom:

            out["values"] = self.values.tolist()

        else:

            out["vmin"] = self.vmin
            out["vmax"] = self.vmax
            out["scale"] = self.scale
            out["n_points"] = self.n_points

        return out


@dataclass(frozen=True)
class ParameterGrid:

    parameter_list: List[Parameter]
    energy_grid: EnergyGrid

    @property
    def n_points(self) -> int:

        n = 1
        for param in self.parameter_list:

            n *= len(param.grid)

        return n

    @property
    def n_parameters(self) -> int:

        return len(self.parameter_list)

    @classmethod
    def from_dict(cls, d: Dict[str, Dict[str, Any]]) -> "ParameterGrid":

        # make sure to sort so that we always have the same parameter
        # ordering

        energy_grid = EnergyGrid.from_dict(d.pop("energy_grid"))

        pars = list(d.keys())

        pars.sort()

        par_list = [
            Parameter.from_dict(par_name, d[par_name]) for par_name in pars
        ]

        return cls(par_list, energy_grid)

    @classmethod
    def from_yaml(cls, file_name: str) -> "ParameterGrid":

        with open(file_name, 'r') as f:

            inputs = yaml.load(stream=f, Loader=yaml.SafeLoader)

        return cls.from_dict(inputs)

    @property
    def parameter_names(self) -> List[str]:

        return [p.name for p in self.parameter_list]

    def to_dict(self) -> Dict[str, Dict[str, Any]]:

        out = {}

        for p in self.parameter_list:

            out[p.name] = p.to_dict()

        out['energy_grid'] = self.energy_grid.to_dict()

        return out

    def write(self, file_name: str) -> None:

        with open(file_name, "w") as f:

            yaml.dump(
                stream=f,
                data=self.to_dict(),
                default_flow_style=False,
                Dumper=yaml.SafeDumper,
            )

    # def to_hdf5_group(self, f: h5py.File) -> None:

    #     recursively_save_dict_contents_to_group(
    #         f, "parameter_grid", self.to_dict()
    #     )

    def at_index(self, i: int) -> Dict[str, float]:
        """
        return the ith set of parameters

        :param i:
        :type i: int
        :returns:

        """
        idx = 0

        for result in itertools.product(*[p.grid for p in self.parameter_list]):

            if i == idx:

                d = OrderedDict()

                for k, v in zip(self.parameter_names, result):

                    d[k] = v

                return d

            else:

                idx += 1


class SimulationBuilder:
    def __init__(
        self,
        parameter_grid: ParameterGrid,
        out_file: str,
        import_line: str,
        n_cores: int = 1,
        n_nodes: Optional[int] = None,
        linear_execution: bool = False,
    ):

        """
        The simulation builder class constructs the scripts
        needed for building the table model database

        :param parameter_grid:
        :type parameter_grid: ParameterGrid
        :param out_file:
        :type out_file: str
        :param import_line:
        :type import_line: str
        :param n_cores:
        :type n_cores: int
        :param n_nodes:
        :type n_nodes: Optional[int]
        :param linear_execution:
        :type linear_execution: bool
        :returns:

        """
        self._import_line: str = import_line

        self._n_cores: int = n_cores

        self._n_nodes: Optional[int] = n_nodes

        self._out_file: str = out_file

        self._base_dir: Path = Path(out_file).parent.absolute()

        self._linear_execution: bool = linear_execution

        # write out the parameter file

        self._parameter_file: Path = self._base_dir / "parameters.yml"

        parameter_grid.write(str(self._parameter_file))

        self._generate_python_script()

    def _generate_python_script(self) -> None:

        py_gen: PythonGenerator = PythonGenerator(
            "run_simulation.py",
            self._out_file,
            str(self._parameter_file),
            self._import_line,
            self._n_cores,
            self._n_nodes,
            self._linear_execution,
        )

        py_gen.write(str(self._base_dir))

    def _generate_slurm_script(self) -> None:

        pass



from pathlib import Path
from typing import Optional


class ScriptGenerator:
    def __init__(self, file_name: str) -> None:

        self._file_name: str = file_name
        self._output: str = ""
        self._build_script()

    def _build_script(self) -> None:
        pass

    def _add_line(self, line: str, indent_level: int = 0) -> None:

        for i in range(indent_level):

            self._output += "\t"

        # add the line

        self._output += line

        # close the line
        self._end_line()

    def _end_line(self):

        self._output += "\n"

    def write(self, directory: str = ".") -> None:

        out_file: Path = Path(directory) / self._file_name

        with out_file.open("w") as f:

            f.write(self._output)


class PythonGenerator(ScriptGenerator):
    def __init__(
        self,
        file_name: str,
        database_file: str,
        parameter_file: str,
        import_line: str,
        n_procs: int,
        n_nodes: Optional[int] = None,
        linear_exceution: bool = False,
    ) -> None:

        self._import_line = import_line
        self._n_procs: int = n_procs
        self._n_nodes: Optional[int] = n_nodes
        self._parameter_file: str = parameter_file
        self._database_file: str = Path(database_file).absolute()
        self._linear_execution: bool = linear_exceution

        super().__init__(file_name)

    def _build_script(self) -> None:

        self._add_line(self._import_line)
        self._add_line("from joblib import Parallel, delayed")
        self._add_line("from ronswanson import ParameterGrid")
        self._end_line()

        self._add_line(
            f"pg = ParameterGrid.from_yaml('{self._parameter_file}')"
        )

        self._add_line("def func(i):")
        self._add_line("params = pg.at_index(i)", indent_level=1)
        self._add_line(
            f"simulation = Simulation(i, params, pg.energy_grid.grid,'{self._database_file}')",
            indent_level=1,
        )
        self._add_line("simulation.run()", indent_level=1)

        if self._n_nodes is None:

            self._add_line("iteration = [i for i in range(0, pg.n_points)]")

        else:

            pass

        if self._linear_execution:

            # just do a straight for loop

            self._add_line("for i in iteration:")
            self._add_line("func(i)", indent_level=1)

        else:

            # use joblib

            self._add_line(
                f"Parallel(n_jobs={self._n_procs})(delayed(func)(i) for i in iteration)"
            )


class SLURMGenerator(ScriptGenerator):
    def __init__(self, file_name: str) -> None:
        super().__init__(file_name)



# -*- coding: utf-8 -*-

"""Top-level package for ronswanson."""

__author__ = """J. Michael Burgess"""
__email__ = 'jburgess@mpe.mpg.de'

from ._version import get_versions
from .database import Database
from .simulation import Simulation
from .simulation_builder import ParameterGrid, SimulationBuilder
from .utils.logging import update_logging_level

__all__ = [
    "Simulation",
    "ParameterGrid",
    "SimulationBuilder",
    "Database",
    "update_logging_level",
]


__version__ = get_versions()['version']
del get_versions
